#include <stdio.h>
#include <time.h>
#include <stdlib.h>
#include <fcntl.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <getopt.h>

#define MAXSLOTS 4096		/* Parallel scan slots */
#define NRETRIES 3		/* How many times re-transmit packet */
#define NTIME 4			/* Delay between packet re-transmits */

extern int udp_port;

struct
{
  struct in_addr addr;
  int inuse;
  int ntry;
  int time;
}
slot[MAXSLOTS];
int slotsinuse;

int nretries = NRETRIES;
int ntime = NTIME;
int outgoing_port;
int maxslots = MAXSLOTS;

int addr_start;
int addr_counter;
int addr_bits;
struct in_addr addr_in_addr;
int verbosity = 1;

struct in_addr *
getnextip ()
{
  if (addr_counter >= 1 << (32 - addr_bits))
    return NULL;

  addr_in_addr.s_addr = htonl (addr_start + addr_counter);
  addr_counter++;
  return &addr_in_addr;
}

int
initslot (int i)
{
  struct in_addr *addr;

  slot[i].inuse = 0;

  addr = getnextip ();
  if (addr == NULL)
    return 0;

  slotsinuse++;
  slot[i].addr.s_addr = addr->s_addr;
  slot[i].ntry = nretries;
  slot[i].time = 0;
  slot[i].inuse = 1;

  return 1;
}

void
usage (char *argv0)
{
  printf ("%s [OPTIONS] [address]\n", argv0);
  printf ("%s [OPTIONS] [network]/[bits]\n", argv0);
  printf ("\n");
  printf ("OPTIONS: \n");
  printf ("-p [port]    Port to connect to (default %d)\n", udp_port);
  printf ("-o [port]    Outgoing port (default: dynamically assigned)\n", udp_port);
  printf ("-r [retr]    How many times retransmit packet (default %d)\n", nretries);
  printf ("-t [seconds] Delay between packet re-transmits (default %d secs)\n", ntime);
  printf ("-s [hosts]   Number of scanned hosts in parallel (maximum/default: %d)\n", maxslots);
  printf ("\n");
  printf ("Copyright (c) 2002 Martin Hinner, SECONS Ltd\n");
  printf ("Send bug reports to bugs@secons.com\n");
  exit (0);
}

int format_packet (void *buffer, int maxlen, struct in_addr addr);
int received_packet (void *buffer, int len, struct in_addr addr);


int
main (int argc, char **argv)
{
  struct sockaddr_in sin;
  int sock;
  int i;
  char *s, *argv0;
  fd_set readfds;
  struct timeval tv;
  int len, datalen;
  int tim;
  char buffer[1024];
  int sentpkt;
  int rc;
  char ch;

  argv0 = argv[0];

  while ((ch = getopt (argc, argv, "s:r:t:p:o:vqh")) != -1)
    switch (ch)
      {
      case 's':
         maxslots = atoi(optarg);
         if (maxslots > MAXSLOTS) {
           fprintf(stderr,"Invalid value %d ignored, using %d\n", maxslots, MAXSLOTS);
         }
         break;
      case 'r':
         nretries = atoi(optarg);
         break;
      case 't':
         ntime = atoi(optarg);
         break;

      case 'o':
         outgoing_port = atoi(optarg);
         break;
      case 'p':
         udp_port = atoi(optarg);
         break;
      case 'q':
         verbosity--;
         break;
      case 'v':
         verbosity++;
         break;
      case 'h':
        usage (argv0);
        exit (1);
      }

  argc -= optind;
  argv += optind;


  if (argc != 1)
    {
      usage (argv0);
    }

  s = argv[0];
  while (*s && *s != '/')
    s++;
  if (*s == 0)
    addr_bits = 32;
  else
    {
      *s++ = 0;
      addr_bits = atoi (s);
    }
  inet_aton (argv[0], &addr_in_addr);
  addr_start = ntohl (addr_in_addr.s_addr) & (~((1 << (32 - addr_bits)) - 1));
  addr_start = htonl (addr_start);
  if (verbosity >= 1)
    fprintf (stderr, "** Scanning %s/%u...\n",
	     inet_ntoa (*(struct in_addr *) &addr_start), addr_bits);
  addr_start = ntohl (addr_start);

  sock = socket (AF_INET, SOCK_DGRAM, 17);
  if (sock < 0)
    {
      perror ("socket");
      exit (1);
    }


  sin.sin_family = AF_INET;
  sin.sin_addr.s_addr = htonl (INADDR_ANY);
  sin.sin_port = htons (outgoing_port);

  if (bind (sock, (struct sockaddr *) &sin, sizeof (sin)) < 0)
    {
      perror ("bind");
      exit (1);
    }

  for (i = 0; i < maxslots; i++)
    initslot (i);

  while (slotsinuse)
    {
      tim = time (NULL);

      sentpkt = 0;
      for (i = 0; i < maxslots && sentpkt == 0; i++)
	if (slot[i].inuse)
	  {
	    if (slot[i].time < tim - ntime)
	      {
		if (slot[i].ntry > 0)
		  {
		    slot[i].ntry--;
		    slot[i].time = tim;

		    sin.sin_family = AF_INET;
		    sin.sin_addr.s_addr = slot[i].addr.s_addr;
		    sin.sin_port = htons (udp_port);

		    datalen =
		      format_packet (buffer, sizeof (buffer), slot[i].addr);
		    if (sendto
			(sock, buffer, datalen, 0, (struct sockaddr *) &sin,
			 sizeof (sin)) == datalen)
		      sentpkt++;
		    else
		      {
			if (verbosity >= 1)
			  printf ("%s <unreachable>\n",
				  inet_ntoa (slot[i].addr));
			fflush (stdout);
			slotsinuse--;
			initslot (i);
		      }
		  }
		else
		  {
		    if (verbosity >= 1)
		      printf ("%s <timeout>\n", inet_ntoa (slot[i].addr));
		    fflush (stdout);
		    slotsinuse--;
		    initslot (i);
		  }
	      }
	  }

      tv.tv_sec = 0;
      tv.tv_usec = sentpkt ? 0 : 200000;

      FD_ZERO (&readfds);
      FD_SET (0, &readfds);
      FD_SET (sock, &readfds);

      select (255, &readfds, NULL, NULL, &tv);

      if (FD_ISSET (sock, &readfds))
	{
	  len = sizeof (sin);
	  rc = recvfrom (sock, buffer, sizeof (buffer),
			 0, (struct sockaddr *) &sin, &len);
	  if (rc > 0)
	    {
	      for (i = 0; i < maxslots; i++)
		{
		  if (slot[i].inuse
		      && slot[i].addr.s_addr == sin.sin_addr.s_addr)
		    {
		      if (received_packet (buffer, rc, sin.sin_addr) > 0)
			{
			  fflush (stdout);
			  slotsinuse--;
			  initslot (i);
			}
		    }
		}
	    }
	}			/* if FD_ISSET */

    }				/* while slotsinuse */

  return 0;
}
